import inspect
import os

from collections.abc import Callable
from prometheus_client import start_wsgi_server, REGISTRY, CollectorRegistry

from .prometheus_url import Generator


def get_module_name(func: Callable) -> str:
    """Get the name of the module that contains the function."""
    module = inspect.getmodule(func)
    if module is None or module.__name__ == "__main__":
        return get_filename_as_module(func)
    return module.__name__


def get_filename_as_module(func: Callable) -> str:
    """Get the filename of the module that contains the function."""
    fullpath = inspect.getsourcefile(func)
    if fullpath is None:
        return ""

    filename = os.path.basename(fullpath)
    module_part = os.path.splitext(filename)[0]
    return module_part


def get_function_name(func: Callable) -> str:
    """Get the name of the function."""
    return func.__qualname__ or func.__name__


def write_docs(func_name: str, module_name: str):
    """Write the prometheus query urls to the function docstring."""
    generator = Generator(func_name, module_name)
    docs = f"Prometheus Query URLs for Function - {func_name} and Module - {module_name}: \n\n"

    urls = generator.create_urls()
    for key, value in urls.items():
        docs = f"{docs}{key} : {value} \n\n"

    docs = f"{docs}-------------------------------------------\n"
    return docs


def append_docs_to_docstring(func, func_name, module_name):
    """Helper for appending docs to a function's docstring."""
    if func.__doc__ is None:
        return write_docs(func_name, module_name)
    else:
        return f"{func.__doc__}\n{write_docs(func_name, module_name)}"


def start_http_server(
    port: int = 9464, addr: str = "0.0.0.0", registry: CollectorRegistry = REGISTRY
):
    """Starts a WSGI server for prometheus metrics as a daemon thread."""
    start_wsgi_server(port, addr, registry)



from typing_extensions import Unpack

from .settings import AutometricsOptions, init_settings
from .tracker import set_tracker, get_tracker_type


def init(**kwargs: Unpack[AutometricsOptions]):
    """Optional initialization function to be used instead of setting environment variables. You can override settings by calling init with the new settings."""
    init_settings(**kwargs)
    tracker_type = get_tracker_type()
    set_tracker(tracker_type)



from enum import Enum
from typing import Optional, Tuple


class ObjectivePercentile(Enum):
    """The percentage of requests that must meet the given criteria (success rate or latency)."""

    P90 = "90"
    P95 = "95"
    P99 = "99"
    P99_9 = "99.9"


class ObjectiveLatency(Enum):
    """The latency threshold for the given percentile."""

    Ms5 = "0.005"
    Ms10 = "0.01"
    Ms25 = "0.025"
    Ms50 = "0.05"
    Ms75 = "0.075"
    Ms100 = "0.1"
    Ms250 = "0.25"
    Ms500 = "0.5"
    Ms750 = "0.75"
    Ms1000 = "1.0"
    Ms2500 = "2.5"
    Ms5000 = "5.0"
    Ms7500 = "7.5"
    Ms10000 = "10.0"


# This represents a Service-Level Objective (SLO) for a function or group of functions.
# The objective should be given a descriptive name and can represent
# a success rate and/or latency objective.
#
# For details on SLOs, see <https://sre.google/sre-book/service-level-objectives/>
#
# Example:
# ```python
# from autometrics import autometrics
# from autometrics.objectives import Objective, ObjectivePercentile, ObjectiveLatency
# API_SLO = Objective(
#     "api",
#     success_rate=ObjectivePercentile.P99_9,
#     latency=(ObjectiveLatency.Ms250, ObjectivePercentile.P99),
# )
#
# @autometrics(objective = API_SLO)
# def api_handler() :
#    # ...
# ```
#
# ## How this works
#
# When an objective is added to a function, the metrics for that function will
# have additional labels attached to specify the SLO details.
#
# Autometrics comes with a set of Prometheus [recording rules](https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/)
# and [alerting rules](https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/)
# that will fire alerts when the given objective is being violated.
#
# By default, these recording rules will effectively lay dormant.
# However, they are enabled when the special labels are present on certain metrics.
class Objective:
    """A Service-Level Objective (SLO) for a function or group of functions."""

    name: str
    """name: The name of the objective. This should be something descriptive of the function or group of functions it covers."""
    success_rate: Optional[ObjectivePercentile]
    """Specify the success rate for this objective.

    This means that the function or group of functions that are part of this objective
    should return an `Ok` result at least this percentage of the time."""
    latency: Optional[Tuple[ObjectiveLatency, ObjectivePercentile]]

    def __init__(
        self,
        name: str,
        success_rate: Optional[ObjectivePercentile] = None,
        latency: Optional[Tuple[ObjectiveLatency, ObjectivePercentile]] = None,
    ):
        """Create a new objective with the given name.

        The name should be something descriptive of the function or group of functions it covers.
        For example, if you have an objective covering all of the HTTP handlers in your API you might call it "api".
        """

        self.name = name
        self.success_rate = success_rate
        self.latency = latency



import time
from typing import Optional
from prometheus_client import Counter, Histogram, Gauge

from ..constants import (
    COUNTER_NAME_PROMETHEUS,
    HISTOGRAM_NAME_PROMETHEUS,
    CONCURRENCY_NAME_PROMETHEUS,
    SERVICE_NAME_PROMETHEUS,
    BUILD_INFO_NAME,
    COUNTER_DESCRIPTION,
    HISTOGRAM_DESCRIPTION,
    CONCURRENCY_DESCRIPTION,
    BUILD_INFO_DESCRIPTION,
    OBJECTIVE_NAME_PROMETHEUS,
    OBJECTIVE_PERCENTILE_PROMETHEUS,
    OBJECTIVE_LATENCY_THRESHOLD_PROMETHEUS,
    COMMIT_KEY,
    VERSION_KEY,
    BRANCH_KEY,
)

from .exemplar import get_exemplar
from .tracker import Result
from ..objectives import Objective
from ..settings import get_settings


class PrometheusTracker:
    """A tracker for Prometheus metrics."""

    prom_counter = Counter(
        COUNTER_NAME_PROMETHEUS,
        COUNTER_DESCRIPTION,
        [
            "function",
            "module",
            SERVICE_NAME_PROMETHEUS,
            "result",
            "caller_module",
            "caller_function",
            OBJECTIVE_NAME_PROMETHEUS,
            OBJECTIVE_PERCENTILE_PROMETHEUS,
        ],
    )
    prom_histogram = Histogram(
        HISTOGRAM_NAME_PROMETHEUS,
        HISTOGRAM_DESCRIPTION,
        [
            "function",
            "module",
            SERVICE_NAME_PROMETHEUS,
            OBJECTIVE_NAME_PROMETHEUS,
            OBJECTIVE_PERCENTILE_PROMETHEUS,
            OBJECTIVE_LATENCY_THRESHOLD_PROMETHEUS,
        ],
        buckets=get_settings()["histogram_buckets"],
        unit="seconds",
    )
    prom_gauge_build_info = Gauge(
        BUILD_INFO_NAME,
        BUILD_INFO_DESCRIPTION,
        [COMMIT_KEY, VERSION_KEY, BRANCH_KEY, SERVICE_NAME_PROMETHEUS],
    )
    prom_gauge_concurrency = Gauge(
        CONCURRENCY_NAME_PROMETHEUS,
        CONCURRENCY_DESCRIPTION,
        [
            "function",
            "module",
            SERVICE_NAME_PROMETHEUS,
        ],
    )

    def __init__(self) -> None:
        self._has_set_build_info = False

    def _count(
        self,
        func_name: str,
        module_name: str,
        caller_module: str,
        caller_function: str,
        objective: Optional[Objective] = None,
        exemplar: Optional[dict] = None,
        result: Result = Result.OK,
        inc_by: int = 1,
    ):
        """Increment the counter for the function call."""
        objective_name = "" if objective is None else objective.name
        percentile = (
            ""
            if objective is None or objective.success_rate is None
            else objective.success_rate.value
        )
        service_name = get_settings()["service_name"]

        self.prom_counter.labels(
            func_name,
            module_name,
            service_name,
            result.value,
            caller_module,
            caller_function,
            objective_name,
            percentile,
        ).inc(inc_by, exemplar)

    def _histogram(
        self,
        func_name: str,
        module_name: str,
        start_time: float,
        objective: Optional[Objective] = None,
        exemplar: Optional[dict] = None,
    ):
        """Observe the duration of the function call."""
        duration = time.time() - start_time

        objective_name = "" if objective is None else objective.name
        latency = None if objective is None else objective.latency
        percentile = ""
        threshold = ""
        if latency is not None:
            threshold = latency[0].value
            percentile = latency[1].value
        service_name = get_settings()["service_name"]

        self.prom_histogram.labels(
            func_name,
            module_name,
            service_name,
            objective_name,
            percentile,
            threshold,
        ).observe(duration, exemplar)

    def set_build_info(self, commit: str, version: str, branch: str):
        if not self._has_set_build_info:
            self._has_set_build_info = True
            service_name = get_settings()["service_name"]
            self.prom_gauge_build_info.labels(
                commit, version, branch, service_name
            ).set(1)

    def start(
        self, function: str, module: str, track_concurrency: Optional[bool] = False
    ):
        """Start tracking metrics for a function call."""
        if track_concurrency:
            service_name = get_settings()["service_name"]
            self.prom_gauge_concurrency.labels(function, module, service_name).inc()

    def finish(
        self,
        start_time: float,
        function: str,
        module: str,
        caller_module: str,
        caller_function: str,
        result: Result = Result.OK,
        objective: Optional[Objective] = None,
        track_concurrency: Optional[bool] = False,
    ):
        """Finish tracking metrics for a function call."""
        exemplar = None
        if get_settings()["enable_exemplars"]:
            exemplar = get_exemplar()

        self._count(
            function,
            module,
            caller_module,
            caller_function,
            objective,
            exemplar,
            result,
        )
        self._histogram(function, module, start_time, objective, exemplar)

        if track_concurrency:
            service_name = get_settings()["service_name"]
            self.prom_gauge_concurrency.labels(function, module, service_name).dec()

    def initialize_counters(
        self,
        function: str,
        module: str,
        objective: Optional[Objective] = None,
    ):
        """Initialize tracking metrics for a function call at zero."""
        caller_module = ""
        caller_function = ""
        self._count(
            function,
            module,
            caller_module,
            caller_function,
            objective,
            None,
            Result.OK,
            0,
        )
        self._count(
            function,
            module,
            caller_module,
            caller_function,
            objective,
            None,
            Result.ERROR,
            0,
        )



import time
from typing import Optional

from opentelemetry.metrics import (
    Meter,
    Counter,
    Histogram,
    UpDownCounter,
    set_meter_provider,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.view import View, ExplicitBucketHistogramAggregation
from opentelemetry.exporter.prometheus import PrometheusMetricReader

from .exemplar import get_exemplar
from .tracker import Result
from ..objectives import Objective, ObjectiveLatency
from ..constants import (
    CONCURRENCY_NAME,
    CONCURRENCY_DESCRIPTION,
    COUNTER_DESCRIPTION,
    COUNTER_NAME,
    HISTOGRAM_DESCRIPTION,
    HISTOGRAM_NAME,
    BUILD_INFO_NAME,
    BUILD_INFO_DESCRIPTION,
    SERVICE_NAME,
    OBJECTIVE_NAME,
    OBJECTIVE_PERCENTILE,
    OBJECTIVE_LATENCY_THRESHOLD,
)
from ..settings import get_settings


class OpenTelemetryTracker:
    """Tracker for OpenTelemetry."""

    __counter_instance: Counter
    __histogram_instance: Histogram
    __up_down_counter_build_info_instance: UpDownCounter
    __up_down_counter_concurrency_instance: UpDownCounter

    def __init__(self):
        exporter = PrometheusMetricReader("")
        view = View(
            name=HISTOGRAM_NAME,
            description=HISTOGRAM_DESCRIPTION,
            instrument_name=HISTOGRAM_NAME,
            aggregation=ExplicitBucketHistogramAggregation(
                boundaries=get_settings()["histogram_buckets"]
            ),
        )
        meter_provider = MeterProvider(metric_readers=[exporter], views=[view])
        set_meter_provider(meter_provider)
        meter = meter_provider.get_meter(name="autometrics")
        self.__counter_instance = meter.create_counter(
            name=COUNTER_NAME, description=COUNTER_DESCRIPTION
        )
        self.__histogram_instance = meter.create_histogram(
            name=HISTOGRAM_NAME,
            description=HISTOGRAM_DESCRIPTION,
            unit="seconds",
        )
        self.__up_down_counter_build_info_instance = meter.create_up_down_counter(
            name=BUILD_INFO_NAME,
            description=BUILD_INFO_DESCRIPTION,
        )
        self.__up_down_counter_concurrency_instance = meter.create_up_down_counter(
            name=CONCURRENCY_NAME,
            description=CONCURRENCY_DESCRIPTION,
        )
        self._has_set_build_info = False

    def __count(
        self,
        function: str,
        module: str,
        caller_module: str,
        caller_function: str,
        objective: Optional[Objective],
        exemplar: Optional[dict],
        result: Result,
        inc_by: int = 1,
    ):
        objective_name = "" if objective is None else objective.name
        percentile = (
            ""
            if objective is None or objective.success_rate is None
            else objective.success_rate.value
        )
        self.__counter_instance.add(
            inc_by,
            attributes={
                "function": function,
                "module": module,
                "result": result.value,
                "caller.module": caller_module,
                "caller.function": caller_function,
                OBJECTIVE_NAME: objective_name,
                OBJECTIVE_PERCENTILE: percentile,
                SERVICE_NAME: get_settings()["service_name"],
            },
        )

    def __histogram(
        self,
        function: str,
        module: str,
        start_time: float,
        objective: Optional[Objective],
        exemplar: Optional[dict],
    ):
        duration = time.time() - start_time

        objective_name = "" if objective is None else objective.name
        latency = None if objective is None else objective.latency
        percentile = ""
        threshold = ""

        if latency is not None:
            threshold = latency[0].value
            percentile = latency[1].value

        self.__histogram_instance.record(
            duration,
            attributes={
                "function": function,
                "module": module,
                SERVICE_NAME: get_settings()["service_name"],
                OBJECTIVE_NAME: objective_name,
                OBJECTIVE_PERCENTILE: percentile,
                OBJECTIVE_LATENCY_THRESHOLD: threshold,
            },
        )

    def set_build_info(self, commit: str, version: str, branch: str):
        if not self._has_set_build_info:
            self._has_set_build_info = True
            self.__up_down_counter_build_info_instance.add(
                1.0,
                attributes={
                    "commit": commit,
                    "version": version,
                    "branch": branch,
                    SERVICE_NAME: get_settings()["service_name"],
                },
            )

    def start(
        self,
        function: str,
        module: str,
        track_concurrency: Optional[bool] = False,
    ):
        """Start tracking metrics for a function call."""
        if track_concurrency:
            self.__up_down_counter_concurrency_instance.add(
                1.0,
                attributes={
                    "function": function,
                    "module": module,
                    SERVICE_NAME: get_settings()["service_name"],
                },
            )

    def finish(
        self,
        start_time: float,
        function: str,
        module: str,
        caller_module: str,
        caller_function: str,
        result: Result = Result.OK,
        objective: Optional[Objective] = None,
        track_concurrency: Optional[bool] = False,
    ):
        """Finish tracking metrics for a function call."""

        exemplar = None
        # Currently, exemplars are only supported by prometheus-client
        # https://github.com/autometrics-dev/autometrics-py/issues/41
        # if get_settings()["exemplars"]:
        #     exemplar = get_exemplar()
        self.__count(
            function,
            module,
            caller_module,
            caller_function,
            objective,
            exemplar,
            result,
        )
        self.__histogram(function, module, start_time, objective, exemplar)
        if track_concurrency:
            self.__up_down_counter_concurrency_instance.add(
                -1.0,
                attributes={
                    "function": function,
                    "module": module,
                    SERVICE_NAME: get_settings()["service_name"],
                },
            )

    def initialize_counters(
        self,
        function: str,
        module: str,
        objective: Optional[Objective] = None,
    ):
        """Initialize tracking metrics for a function call at zero."""
        caller_module = ""
        caller_function = ""
        self.__count(
            function,
            module,
            caller_module,
            caller_function,
            objective,
            None,
            Result.OK,
            0,
        )
        self.__count(
            function,
            module,
            caller_module,
            caller_function,
            objective,
            None,
            Result.ERROR,
            0,
        )



from prometheus_client.exposition import generate_latest
import pytest

from . import TrackerType
from ..decorator import autometrics
from .. import init


@pytest.mark.parametrize("tracker", TrackerType)
def test_metrics_format(tracker):
    """Test that the metrics are formatted correctly."""
    init(tracker=tracker.value)

    @autometrics
    def test_function():
        pass

    test_function()

    blob = generate_latest()
    assert blob is not None
    data = blob.decode("utf-8")

    assert "function_calls_total{" in data
    assert "function_calls_duration_seconds_bucket{" in data



from prometheus_client.exposition import generate_latest
import asyncio
import pytest


from .tracker import set_tracker, TrackerType

from ..decorator import autometrics
from ..utils import get_function_name, get_module_name


@autometrics(track_concurrency=True)
async def sleep(time):
    await asyncio.sleep(time)


@pytest.mark.asyncio
async def test_concurrency_tracking_prometheus(monkeypatch):
    # HACK - We need to set the tracker explicitly here, instead of using `init_tracker`
    #        because the library was already initialized with the OpenTelemetry tracker
    set_tracker(TrackerType.PROMETHEUS)

    func_name = get_function_name(sleep)
    module_name = get_module_name(sleep)

    # Create a 200ms async task
    loop = asyncio.get_event_loop()
    task = loop.create_task(sleep(0.2))

    # Await a separate 100ms async task.
    # This way, the 200ms task will still running once this task is done.
    # We have to do this to ensure that the 200ms task is kicked off before we call `generate_latest`
    await sleep(0.1)
    blob = generate_latest()
    await task
    assert blob is not None
    data = blob.decode("utf-8")

    assert (
        f"""# TYPE function_calls_concurrent gauge\nfunction_calls_concurrent{{function="sleep",module="autometrics.tracker.test_concurrency",service_name="autometrics"}} 1.0"""
        in data
    )


# NOTE - Gauges are not supported by the OTEL collector (yet)
#        https://github.com/open-telemetry/opentelemetry-python/pull/3306
#
# @pytest.mark.asyncio
# async def test_concurrency_tracking_opentelemetry(monkeypatch):
#     # HACK - We need to set the tracker explicitly here, instead of using `init_tracker`
#     #        because the library was already initialized with the OpenTelemetry tracker
#     set_tracker(TrackerType.OPENTELEMETRY)

#     # Create a 200ms async task
#     loop = asyncio.get_event_loop()
#     task = loop.create_task(sleep(0.2))

#     # Await a separate 100ms async task.
#     # This way, the 200ms task will still running once this task is done.
#     # We have to do this to ensure that the 200ms task is kicked off before we call `generate_latest`
#     await sleep(0.1)
#     blob = generate_latest()
#     await task
#     assert blob is not None
#     data = blob.decode("utf-8")
#     print(data)
#     assert (
#         f"""# TYPE function_calls_concurrent gauge\nfunction_calls_concurrent{{function="sleep",module="test_concurrency"}} 1.0"""
#         in data
#     )



"""Tests for caller tracking."""
from functools import wraps
from prometheus_client.exposition import generate_latest

from .decorator import autometrics


def test_caller_detection():
    """This is a test to see if the caller is properly detected."""

    def dummy_decorator(func):
        @wraps(func)
        def dummy_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return dummy_wrapper

    def another_decorator(func):
        @wraps(func)
        def another_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return another_wrapper

    @dummy_decorator
    @autometrics
    @another_decorator
    def foo():
        pass

    @autometrics
    def bar():
        foo()

    bar()

    blob = generate_latest()
    assert blob is not None
    data = blob.decode("utf-8")

    expected = """function_calls_total{caller_function="test_caller_detection.<locals>.bar",caller_module="autometrics.test_caller",function="test_caller_detection.<locals>.foo",module="autometrics.test_caller",objective_name="",objective_percentile="",result="ok",service_name="autometrics"} 1.0"""
    assert "wrapper" not in data
    assert expected in data



import urllib.parse
import os
from typing import Optional
from dotenv import load_dotenv

ADD_BUILD_INFO_LABELS = "* on (instance, job) group_left(version, commit) (last_over_time(build_info[1s]) or on (instance, job) up)"


def cleanup_url(url: str) -> str:
    """Remove the trailing slash if there is one."""
    if url[-1] == "/":
        url = url[:-1]
    return url


class Generator:
    """Generate prometheus query urls for a given function/module."""

    def __init__(
        self, function_name: str, module_name: str, base_url: Optional[str] = None
    ):
        load_dotenv()
        self.function_name = function_name
        self.module_name = module_name

        url = base_url or os.getenv("PROMETHEUS_URL") or "http://localhost:9090"
        self.base_url = cleanup_url(url)

    def create_urls(self):
        """Create the prometheus query urls for the function and module."""
        request_rate_query = f'sum by (function, module, commit, version) (rate (function_calls_count_total{{function="{self.function_name}",module="{self.module_name}"}}[5m]) {ADD_BUILD_INFO_LABELS})'
        latency_query = f'sum by (le, function, module, commit, version) (rate(function_calls_duration_bucket{{function="{self.function_name}",module="{self.module_name}"}}[5m]) {ADD_BUILD_INFO_LABELS})'
        error_ratio_query = f'sum by (function, module, commit, version) (rate (function_calls_count_total{{function="{self.function_name}",module="{self.module_name}", result="error"}}[5m]) {ADD_BUILD_INFO_LABELS}) / {request_rate_query}'

        queries = {
            "Request rate URL": request_rate_query,
            "Latency URL": latency_query,
            "Error Ratio URL": error_ratio_query,
        }

        urls = {}
        for [name, query] in queries.items():
            # for query in queries:
            generated_url = self.create_prometheus_url(query)
            urls[name] = generated_url
        return urls

    def create_prometheus_url(self, query: str):
        """Create a the full query url for a given query."""
        encoded_query = urllib.parse.quote(query)
        url = f"{self.base_url}/graph?g0.expr={encoded_query}&g0.tab=0"
        return url



from prometheus_client.exposition import generate_latest
import pytest

from .opentelemetry import OpenTelemetryTracker
from .prometheus import PrometheusTracker

from .tracker import get_tracker, default_tracker

from .. import init


def test_default_tracker(monkeypatch):
    """Test the default tracker type."""

    monkeypatch.delenv("AUTOMETRICS_TRACKER", raising=False)
    init()
    tracker = default_tracker()
    assert isinstance(tracker, OpenTelemetryTracker)

    monkeypatch.setenv("AUTOMETRICS_TRACKER", "prometheus")
    init()
    tracker = default_tracker()
    assert isinstance(tracker, PrometheusTracker)
    monkeypatch.setenv("AUTOMETRICS_TRACKER", "PROMETHEUS")
    init()
    tracker = default_tracker()
    assert isinstance(tracker, PrometheusTracker)

    # Should use open telemetry when the tracker is not recognized
    monkeypatch.setenv("AUTOMETRICS_TRACKER", "something_else")
    init()
    tracker = default_tracker()
    assert isinstance(tracker, OpenTelemetryTracker)


def test_init_prometheus_tracker_set_build_info(monkeypatch):
    """Test that init_tracker (for a Prometheus tracker) calls set_build_info using env vars."""

    commit = "d6abce3"
    version = "1.0.1"
    branch = "main"
    tracker = "prometheus"

    monkeypatch.setenv("AUTOMETRICS_COMMIT", commit)
    monkeypatch.setenv("AUTOMETRICS_VERSION", version)
    monkeypatch.setenv("AUTOMETRICS_BRANCH", branch)
    monkeypatch.setenv("AUTOMETRICS_TRACKER", tracker)
    init()

    prom_tracker = get_tracker()
    assert isinstance(prom_tracker, PrometheusTracker)

    blob = generate_latest()
    assert blob is not None
    data = blob.decode("utf-8")
    print(data)

    prom_build_info = f"""build_info{{branch="{branch}",commit="{commit}",service_name="autometrics",version="{version}"}} 1.0"""
    assert prom_build_info in data

    monkeypatch.delenv("AUTOMETRICS_VERSION", raising=False)
    monkeypatch.delenv("AUTOMETRICS_COMMIT", raising=False)
    monkeypatch.delenv("AUTOMETRICS_BRANCH", raising=False)


def test_init_otel_tracker_set_build_info(monkeypatch):
    """
    Test that init_tracker (for an OTEL tracker) calls set_build_info using env vars.
    Note that the OTEL collector translates metrics to Prometheus.
    """
    pytest.skip(
        "Skipping test because OTEL collector does not create a gauge when it translates UpDownCounter to Prometheus"
    )

    commit = "a29a178"
    version = "0.0.1"
    branch = "main"

    monkeypatch.setenv("AUTOMETRICS_COMMIT", commit)
    monkeypatch.setenv("AUTOMETRICS_VERSION", version)
    monkeypatch.setenv("AUTOMETRICS_BRANCH", branch)
    init()

    otel_tracker = get_tracker()
    assert isinstance(otel_tracker, OpenTelemetryTracker)

    blob = generate_latest()
    assert blob is not None
    data = blob.decode("utf-8")

    prom_build_info = f"""build_info{{branch="{branch}",commit="{commit}",service_name="autometrics",version="{version}",service_name="autometrics"}} 1.0"""
    assert prom_build_info in data

    monkeypatch.delenv("AUTOMETRICS_VERSION", raising=False)
    monkeypatch.delenv("AUTOMETRICS_COMMIT", raising=False)
    monkeypatch.delenv("AUTOMETRICS_BRANCH", raising=False)



from opentelemetry import trace


def get_exemplar():
    """Generates an exemplar dictionary from the current implicit OTel context if available"""
    span_context = trace.get_current_span().get_span_context()

    # Only include the exemplar if it is valid and sampled
    if span_context.is_valid and span_context.trace_flags.sampled:
        # You must set the trace_id and span_id exemplar labels like this to link OTel and
        # Prometheus. They must be formatted as hexadecimal strings.
        return {
            "trace_id": trace.format_trace_id(span_context.trace_id),
            "span_id": trace.format_span_id(span_context.span_id),
        }

    return None



from typing import Protocol, Optional
from enum import Enum

from ..objectives import Objective
from ..settings import get_settings


class Result(Enum):
    """Result of the function call."""

    OK = "ok"
    ERROR = "error"


class TrackMetrics(Protocol):
    """Protocol for tracking metrics."""

    def set_build_info(self, commit: str, version: str, branch: str):
        """Observe the build info. Should only be called once per tracker instance"""

    def start(
        self, function: str, module: str, track_concurrency: Optional[bool] = False
    ):
        """Start tracking metrics for a function call."""

    def finish(
        self,
        start_time: float,
        function: str,
        module: str,
        caller_module: str,
        caller_function: str,
        result: Result = Result.OK,
        objective: Optional[Objective] = None,
        track_concurrency: Optional[bool] = False,
    ):
        """Finish tracking metrics for a function call."""

    def initialize_counters(
        self,
        function: str,
        module: str,
        objective: Optional[Objective] = None,
    ):
        """Initialize (counter) metrics for a function at zero."""


class TrackerType(Enum):
    """Type of tracker."""

    OPENTELEMETRY = "opentelemetry"
    PROMETHEUS = "prometheus"


def init_tracker(tracker_type: TrackerType) -> TrackMetrics:
    """Create a tracker"""

    tracker_instance: TrackMetrics
    if tracker_type == TrackerType.OPENTELEMETRY:
        # pylint: disable=import-outside-toplevel
        from .opentelemetry import OpenTelemetryTracker

        tracker_instance = OpenTelemetryTracker()
    elif tracker_type == TrackerType.PROMETHEUS:
        # pylint: disable=import-outside-toplevel
        from .prometheus import PrometheusTracker

        tracker_instance = PrometheusTracker()

    settings = get_settings()
    # NOTE - Only set the build info when the tracker is initialized
    tracker_instance.set_build_info(
        commit=settings["commit"],
        version=settings["version"],
        branch=settings["branch"],
    )

    return tracker_instance


def get_tracker_type() -> TrackerType:
    """Get the tracker type."""
    tracker_type = get_settings()["tracker"]

    if tracker_type.lower() == "prometheus":
        return TrackerType.PROMETHEUS
    return TrackerType.OPENTELEMETRY


def default_tracker():
    """Setup the default tracker."""
    preferred_tracker = get_tracker_type()
    return init_tracker(preferred_tracker)


tracker: TrackMetrics = default_tracker()


def get_tracker() -> TrackMetrics:
    """Get the tracker type."""
    return tracker


def set_tracker(tracker_type: TrackerType):
    """Set the tracker type."""
    global tracker
    tracker = init_tracker(tracker_type)



import pytest

from .prometheus_url import Generator


# Defaults to localhost:9090
@pytest.fixture(name="default_url_generator", autouse=True)
def fixture_default_url() -> Generator:
    """Create a generator that uses the default url."""
    return Generator("myFunction", "myModule")


def test_create_prometheus_url_with_default_url(default_url_generator: Generator):
    """Test that the prometheus url is created correctly."""
    url = default_url_generator.create_prometheus_url("myQuery")

    # Make sure the base URL is correct
    assert url.startswith("http://localhost:9090/graph?g0.expr=")

    # Make sure the query is included in the URL
    assert "myQuery" in url


def test_create_urls_with_default_url(default_url_generator: Generator):
    urls = default_url_generator.create_urls()

    print(urls)

    result = {
        "Request rate URL": "http://localhost:9090/graph?g0.expr=sum%20by%20%28function%2C%20module%2C%20commit%2C%20version%29%20%28rate%20%28function_calls_count_total%7Bfunction%3D%22myFunction%22%2Cmodule%3D%22myModule%22%7D%5B5m%5D%29%20%2A%20on%20%28instance%2C%20job%29%20group_left%28version%2C%20commit%29%20%28last_over_time%28build_info%5B1s%5D%29%20or%20on%20%28instance%2C%20job%29%20up%29%29&g0.tab=0",
        "Latency URL": "http://localhost:9090/graph?g0.expr=sum%20by%20%28le%2C%20function%2C%20module%2C%20commit%2C%20version%29%20%28rate%28function_calls_duration_bucket%7Bfunction%3D%22myFunction%22%2Cmodule%3D%22myModule%22%7D%5B5m%5D%29%20%2A%20on%20%28instance%2C%20job%29%20group_left%28version%2C%20commit%29%20%28last_over_time%28build_info%5B1s%5D%29%20or%20on%20%28instance%2C%20job%29%20up%29%29&g0.tab=0",
        "Error Ratio URL": "http://localhost:9090/graph?g0.expr=sum%20by%20%28function%2C%20module%2C%20commit%2C%20version%29%20%28rate%20%28function_calls_count_total%7Bfunction%3D%22myFunction%22%2Cmodule%3D%22myModule%22%2C%20result%3D%22error%22%7D%5B5m%5D%29%20%2A%20on%20%28instance%2C%20job%29%20group_left%28version%2C%20commit%29%20%28last_over_time%28build_info%5B1s%5D%29%20or%20on%20%28instance%2C%20job%29%20up%29%29%20/%20sum%20by%20%28function%2C%20module%2C%20commit%2C%20version%29%20%28rate%20%28function_calls_count_total%7Bfunction%3D%22myFunction%22%2Cmodule%3D%22myModule%22%7D%5B5m%5D%29%20%2A%20on%20%28instance%2C%20job%29%20group_left%28version%2C%20commit%29%20%28last_over_time%28build_info%5B1s%5D%29%20or%20on%20%28instance%2C%20job%29%20up%29%29&g0.tab=0",
    }
    assert result == urls


@pytest.fixture(name="custom_url_generator", autouse=True)
def fixture_custom_url():
    return Generator("myFunction", "myModule", base_url="http://localhost:9091")


def test_create_prometheus_url_with_custom_url(custom_url_generator: Generator):
    """Test the prometheus url generator with a custom base URL."""
    url = custom_url_generator.create_prometheus_url("myQuery")

    # Make sure the base URL is correct
    assert url.startswith("http://localhost:9091/graph?g0.expr=")

    # Make sure the query is included in the URL
    assert "myQuery" in url



"""Autometrics module."""
import time
import inspect

from contextvars import ContextVar, Token
from functools import wraps
from typing import overload, TypeVar, Callable, Optional, Awaitable
from typing_extensions import ParamSpec

from .objectives import Objective
from .tracker import get_tracker, Result
from .utils import (
    get_function_name,
    get_module_name,
    append_docs_to_docstring,
)


P = ParamSpec("P")
T = TypeVar("T")


caller_module_var: ContextVar[str] = ContextVar("caller.module", default="")
caller_function_var: ContextVar[str] = ContextVar("caller.function", default="")


# Bare decorator usage
@overload
def autometrics(func: Callable[P, T]) -> Callable[P, T]:
    ...


@overload
def autometrics(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    ...


# Decorator with arguments
@overload
def autometrics(
    *, objective: Optional[Objective] = None, track_concurrency: Optional[bool] = False
) -> Callable:
    ...


def autometrics(
    func: Optional[Callable] = None,
    *,
    objective: Optional[Objective] = None,
    track_concurrency: Optional[bool] = False,
):
    """Decorator for tracking function calls and duration. Supports synchronous and async functions."""

    def register_function_info(
        function: str,
        module: str,
    ):
        get_tracker().initialize_counters(
            function=function, module=module, objective=objective
        )

    def track_start(function: str, module: str):
        get_tracker().start(
            function=function, module=module, track_concurrency=track_concurrency
        )

    def track_result_ok(
        start_time: float,
        function: str,
        module: str,
        caller_module: str,
        caller_function: str,
    ):
        get_tracker().finish(
            start_time,
            function=function,
            module=module,
            caller_module=caller_module,
            caller_function=caller_function,
            objective=objective,
            track_concurrency=track_concurrency,
            result=Result.OK,
        )

    def track_result_error(
        start_time: float,
        function: str,
        module: str,
        caller_module: str,
        caller_function: str,
    ):
        get_tracker().finish(
            start_time,
            function=function,
            module=module,
            caller_module=caller_module,
            caller_function=caller_function,
            objective=objective,
            track_concurrency=track_concurrency,
            result=Result.ERROR,
        )

    def sync_decorator(func: Callable[P, T]) -> Callable[P, T]:
        """Helper for decorating synchronous functions, to track calls and duration."""

        module_name = get_module_name(func)
        func_name = get_function_name(func)
        register_function_info(func_name, module_name)

        @wraps(func)
        def sync_wrapper(*args: P.args, **kwds: P.kwargs) -> T:
            start_time = time.time()
            caller_module = caller_module_var.get()
            caller_function = caller_function_var.get()
            context_token_module: Optional[Token] = None
            context_token_function: Optional[Token] = None

            try:
                context_token_module = caller_module_var.set(module_name)
                context_token_function = caller_function_var.set(func_name)
                if track_concurrency:
                    track_start(module=module_name, function=func_name)
                result = func(*args, **kwds)
                track_result_ok(
                    start_time,
                    function=func_name,
                    module=module_name,
                    caller_module=caller_module,
                    caller_function=caller_function,
                )

            except Exception as exception:
                result = exception.__class__.__name__
                track_result_error(
                    start_time,
                    function=func_name,
                    module=module_name,
                    caller_module=caller_module,
                    caller_function=caller_function,
                )
                # Reraise exception
                raise exception

            finally:
                if context_token_module is not None:
                    caller_module_var.reset(context_token_module)
                if context_token_function is not None:
                    caller_function_var.reset(context_token_function)

            return result

        sync_wrapper.__doc__ = append_docs_to_docstring(func, func_name, module_name)
        return sync_wrapper

    def async_decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        """Helper for decorating async functions, to track calls and duration."""

        module_name = get_module_name(func)
        func_name = get_function_name(func)
        register_function_info(func_name, module_name)

        @wraps(func)
        async def async_wrapper(*args: P.args, **kwds: P.kwargs) -> T:
            start_time = time.time()
            caller_module = caller_module_var.get()
            caller_function = caller_function_var.get()
            context_token_module: Optional[Token] = None
            context_token_function: Optional[Token] = None

            try:
                context_token_module = caller_module_var.set(module_name)
                context_token_function = caller_function_var.set(func_name)
                if track_concurrency:
                    track_start(module=module_name, function=func_name)
                result = await func(*args, **kwds)
                track_result_ok(
                    start_time,
                    function=func_name,
                    module=module_name,
                    caller_module=caller_module,
                    caller_function=caller_function,
                )

            except Exception as exception:
                result = exception.__class__.__name__
                track_result_error(
                    start_time,
                    function=func_name,
                    module=module_name,
                    caller_module=caller_module,
                    caller_function=caller_function,
                )
                # Reraise exception
                raise exception

            finally:
                if context_token_module is not None:
                    caller_module_var.reset(context_token_module)
                if context_token_function is not None:
                    caller_function_var.reset(context_token_function)

            return result

        async_wrapper.__doc__ = append_docs_to_docstring(func, func_name, module_name)
        return async_wrapper

    def pick_decorator(func: Callable) -> Callable:
        """Pick the correct decorator based on the function type."""
        if inspect.iscoroutinefunction(func):
            return async_decorator(func)
        return sync_decorator(func)

    if func is None:
        return pick_decorator
    elif inspect.iscoroutinefunction(func):
        return async_decorator(func)
    else:
        return sync_decorator(func)



"""Constants used by autometrics"""

COUNTER_NAME = "function.calls"
HISTOGRAM_NAME = "function.calls.duration"
CONCURRENCY_NAME = "function.calls.concurrent"
# NOTE - The Rust implementation does not use `build.info`, instead opts for just `build_info`
BUILD_INFO_NAME = "build_info"
SERVICE_NAME = "service.name"


COUNTER_NAME_PROMETHEUS = COUNTER_NAME.replace(".", "_")
HISTOGRAM_NAME_PROMETHEUS = HISTOGRAM_NAME.replace(".", "_")
CONCURRENCY_NAME_PROMETHEUS = CONCURRENCY_NAME.replace(".", "_")
SERVICE_NAME_PROMETHEUS = SERVICE_NAME.replace(".", "_")

COUNTER_DESCRIPTION = "Autometrics counter for tracking function calls"
HISTOGRAM_DESCRIPTION = "Autometrics histogram for tracking function call duration"
CONCURRENCY_DESCRIPTION = "Autometrics gauge for tracking function call concurrency"
BUILD_INFO_DESCRIPTION = (
    "Autometrics info metric for tracking software version and build details"
)

# The following constants are used to create the labels
OBJECTIVE_NAME = "objective.name"
OBJECTIVE_PERCENTILE = "objective.percentile"
OBJECTIVE_LATENCY_THRESHOLD = "objective.latency_threshold"
VERSION_KEY = "version"
COMMIT_KEY = "commit"
BRANCH_KEY = "branch"

# The values are updated to use underscores instead of periods to avoid issues with prometheus.
# A similar thing is done in the rust library, which supports multiple exporters
OBJECTIVE_NAME_PROMETHEUS = OBJECTIVE_NAME.replace(".", "_")
OBJECTIVE_PERCENTILE_PROMETHEUS = OBJECTIVE_PERCENTILE.replace(".", "_")
OBJECTIVE_LATENCY_THRESHOLD_PROMETHEUS = OBJECTIVE_LATENCY_THRESHOLD.replace(".", "_")



import os

from typing import List, TypedDict, Optional
from typing_extensions import Unpack

from .objectives import ObjectiveLatency


class AutometricsSettings(TypedDict):
    """Settings for autometrics."""

    histogram_buckets: List[float]
    tracker: str
    enable_exemplars: bool
    service_name: str
    commit: str
    version: str
    branch: str


class AutometricsOptions(TypedDict, total=False):
    """User supplied overrides for autometrics settings."""

    histogram_buckets: List[float]
    tracker: str
    enable_exemplars: bool
    service_name: str
    commit: str
    version: str
    branch: str


def get_objective_boundaries():
    """Get the objective latency boundaries as float values in seconds (instead of strings)"""
    return list(map(lambda c: float(c.value), ObjectiveLatency))


settings: Optional[AutometricsSettings] = None


def init_settings(**overrides: Unpack[AutometricsOptions]) -> AutometricsSettings:
    config: AutometricsSettings = {
        "histogram_buckets": overrides.get("histogram_buckets")
        or get_objective_boundaries(),
        "tracker": overrides.get("tracker")
        or os.getenv("AUTOMETRICS_TRACKER")
        or "opentelemetry",
        "enable_exemplars": overrides.get(
            "enable_exemplars", os.getenv("AUTOMETRICS_EXEMPLARS") == "true"
        ),
        "service_name": overrides.get(
            "service_name",
            os.getenv(
                "AUTOMETRICS_SERVICE_NAME",
                os.getenv("OTEL_SERVICE_NAME", __package__.rsplit(".", 1)[0]),
            ),
        ),
        "commit": overrides.get(
            "commit", os.getenv("AUTOMETRICS_COMMIT", os.getenv("COMMIT_SHA", ""))
        ),
        "branch": overrides.get(
            "branch", os.getenv("AUTOMETRICS_BRANCH", os.getenv("BRANCH_NAME", ""))
        ),
        "version": overrides.get("version", os.getenv("AUTOMETRICS_VERSION", "")),
    }
    global settings
    settings = config
    return settings


def get_settings() -> AutometricsSettings:
    """Get the current settings."""
    global settings
    if settings is None:
        return init_settings()
    return settings



"""Test the autometrics decorator."""
import time
import asyncio
from prometheus_client.exposition import generate_latest
import pytest

from .decorator import autometrics
from .objectives import ObjectiveLatency, Objective, ObjectivePercentile
from .tracker import set_tracker, TrackerType
from .utils import get_function_name, get_module_name


def basic_function(sleep_duration: float = 0.0):
    """This is a basic function."""
    time.sleep(sleep_duration)
    return True


def error_function():
    """This is a function that raises an error."""
    raise RuntimeError("This is a test error")


async def basic_async_function(sleep_duration: float = 1.0):
    """This is a basic async function."""
    await asyncio.sleep(sleep_duration)
    return True


async def error_async_function():
    """This is an async function that raises an error."""
    await asyncio.sleep(0.5)
    raise RuntimeError("This is a test error")


def never_called_function():
    """This is a sync function that should never be called. Used for testing initialization at zero for counters"""
    raise RuntimeError("This function should never be called")


async def never_called_async_function():
    """This is an async function that should never be called. Used for testing initialization at zero for counters"""
    raise RuntimeError("This function should never be called")


tracker_types = [TrackerType.PROMETHEUS, TrackerType.OPENTELEMETRY]


@pytest.fixture(scope="class", params=tracker_types)
def setup_tracker_type(request):
    """Force the use of a specific metrics tracker"""
    set_tracker(request.param)


@pytest.mark.usefixtures("setup_tracker_type")
class TestDecoratorClass:
    """Test the autometrics decorator. These tests are run multiple times with different trackers."""

    def test_basic(self):
        """This is a basic test."""

        wrapped_function = autometrics(basic_function)
        wrapped_function()

        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count = f"""function_calls_total{{caller_function="",caller_module="",function="basic_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="ok",service_name="autometrics"}} 1.0"""
        assert total_count in data

        for latency in ObjectiveLatency:
            query = f"""function_calls_duration_seconds_bucket{{function="basic_function",le="{latency.value}",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
            assert query in data

        duration_count = f"""function_calls_duration_seconds_count{{function="basic_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_count in data

        duration_sum = f"""function_calls_duration_seconds_sum{{function="basic_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_sum in data

    @pytest.mark.asyncio
    async def test_basic_async(self):
        """This is a basic test."""

        wrapped_function = autometrics(basic_async_function)

        # Test that the function is *still* async after we wrap it
        assert asyncio.iscoroutinefunction(wrapped_function) == True

        await wrapped_function()

        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count = f"""function_calls_total{{caller_function="",caller_module="",function="basic_async_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="ok",service_name="autometrics"}} 1.0"""
        assert total_count in data

        for latency in ObjectiveLatency:
            query = f"""function_calls_duration_seconds_bucket{{function="basic_async_function",le="{latency.value}",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
            assert query in data

        duration_count = f"""function_calls_duration_seconds_count{{function="basic_async_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_count in data

        duration_sum = f"""function_calls_duration_seconds_sum{{function="basic_async_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_sum in data

    def test_objectives(self):
        """This is a test that covers objectives."""

        # set up the function + objective variables
        objective_name = "test_objective"
        success_rate = ObjectivePercentile.P90
        latency = (ObjectiveLatency.Ms100, ObjectivePercentile.P99)
        objective = Objective(
            name=objective_name, success_rate=success_rate, latency=latency
        )
        function_name = get_function_name(basic_function)
        module_name = get_module_name(basic_function)
        wrapped_function = autometrics(objective=objective)(basic_function)

        sleep_duration = 0.25
        # call the function
        wrapped_function(sleep_duration)

        # get the metrics
        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count = f"""function_calls_total{{caller_function="",caller_module="",function="{function_name}",module="autometrics.test_decorator",objective_name="{objective_name}",objective_percentile="{success_rate.value}",result="ok",service_name="autometrics"}} 1.0"""
        assert total_count in data

        # Check the latency buckets
        for objective in ObjectiveLatency:
            count = 0 if float(objective.value) <= sleep_duration else 1
            query = f"""function_calls_duration_seconds_bucket{{function="{function_name}",le="{objective.value}",module="autometrics.test_decorator",objective_latency_threshold="{latency[0].value}",objective_name="{objective_name}",objective_percentile="{latency[1].value}",service_name="autometrics"}} {count}"""
            assert query in data

        duration_count = f"""function_calls_duration_seconds_count{{function="{function_name}",module="autometrics.test_decorator",objective_latency_threshold="{latency[0].value}",objective_name="{objective_name}",objective_percentile="{latency[1].value}",service_name="autometrics"}}"""
        assert duration_count in data

        duration_sum = f"""function_calls_duration_seconds_sum{{function="{function_name}",module="autometrics.test_decorator",objective_latency_threshold="{latency[0].value}",objective_name="{objective_name}",objective_percentile="{latency[1].value}",service_name="autometrics"}}"""
        assert duration_sum in data

    @pytest.mark.asyncio
    async def test_objectives_async(self):
        """This is a test that covers objectives for async functions."""

        # set up the function + objective variables
        objective_name = "test_objective"
        success_rate = ObjectivePercentile.P90
        latency = (ObjectiveLatency.Ms100, ObjectivePercentile.P99)
        objective = Objective(
            name=objective_name, success_rate=success_rate, latency=latency
        )
        wrapped_function = autometrics(objective=objective)(basic_async_function)

        sleep_duration = 0.25

        # Test that the function is *still* async after we wrapped it
        assert asyncio.iscoroutinefunction(wrapped_function) == True

        await wrapped_function(sleep_duration)

        # get the metrics
        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count = f"""function_calls_total{{caller_function="",caller_module="",function="basic_async_function",module="autometrics.test_decorator",objective_name="{objective_name}",objective_percentile="{success_rate.value}",result="ok",service_name="autometrics"}} 1.0"""
        assert total_count in data

        # Check the latency buckets
        for objective in ObjectiveLatency:
            count = 0 if float(objective.value) <= sleep_duration else 1
            query = f"""function_calls_duration_seconds_bucket{{function="basic_async_function",le="{objective.value}",module="autometrics.test_decorator",objective_latency_threshold="{latency[0].value}",objective_name="{objective_name}",objective_percentile="{latency[1].value}",service_name="autometrics"}} {count}"""
            assert query in data

        duration_count = f"""function_calls_duration_seconds_count{{function="basic_async_function",module="autometrics.test_decorator",objective_latency_threshold="{latency[0].value}",objective_name="{objective_name}",objective_percentile="{latency[1].value}",service_name="autometrics"}}"""
        assert duration_count in data

        duration_sum = f"""function_calls_duration_seconds_sum{{function="basic_async_function",module="autometrics.test_decorator",objective_latency_threshold="{latency[0].value}",objective_name="{objective_name}",objective_percentile="{latency[1].value}",service_name="autometrics"}}"""
        assert duration_sum in data

    def test_exception(self):
        """This is a test that covers exceptions."""

        wrapped_function = autometrics(error_function)

        with pytest.raises(RuntimeError) as exception:
            wrapped_function()
        assert "This is a test error" in str(exception.value)

        # get the metrics
        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count = f"""function_calls_total{{caller_function="",caller_module="",function="error_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="error",service_name="autometrics"}} 1.0"""
        assert total_count in data

        for latency in ObjectiveLatency:
            query = f"""function_calls_duration_seconds_bucket{{function="error_function",le="{latency.value}",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
            assert query in data

        duration_count = f"""function_calls_duration_seconds_count{{function="error_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_count in data

        duration_sum = f"""function_calls_duration_seconds_sum{{function="error_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_sum in data

    @pytest.mark.asyncio
    async def test_async_exception(self):
        """This is a test that covers exceptions."""

        wrapped_function = autometrics(error_async_function)

        # Test that the function is *still* async after we wrap it
        assert asyncio.iscoroutinefunction(wrapped_function) == True

        with pytest.raises(RuntimeError) as exception:
            await wrapped_function()
        assert "This is a test error" in str(exception.value)

        # get the metrics
        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count = f"""function_calls_total{{caller_function="",caller_module="",function="error_async_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="error",service_name="autometrics"}} 1.0"""
        assert total_count in data

        for latency in ObjectiveLatency:
            query = f"""function_calls_duration_seconds_bucket{{function="error_async_function",le="{latency.value}",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
            assert query in data

        duration_count = f"""function_calls_duration_seconds_count{{function="error_async_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_count in data

        duration_sum = f"""function_calls_duration_seconds_sum{{function="error_async_function",module="autometrics.test_decorator",objective_latency_threshold="",objective_name="",objective_percentile="",service_name="autometrics"}}"""
        assert duration_sum in data

    def test_initialize_counters_sync(self):
        """This is a test to see if the function calls metric initializes at 0 after invoking the decorator."""

        autometrics(never_called_function)
        # NOTE - Do not call the function! We want to see if we get counter data for it

        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count_ok = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="ok",service_name="autometrics"}} 0.0"""
        assert total_count_ok in data

        total_count_error = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="error",service_name="autometrics"}} 0.0"""
        assert total_count_error in data

    def test_initialize_counters_sync_with_objective(self):
        """This is a test to see if the function calls metric initializes at 0 after invoking the decorator."""

        objective_name = "test_objective"
        success_rate = ObjectivePercentile.P90
        objective = Objective(name=objective_name, success_rate=success_rate)

        autometrics(objective=objective)(never_called_function)
        # NOTE - Do not call the function! We want to see if we get counter data for it

        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count_ok = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_function",module="autometrics.test_decorator",objective_name="{objective_name}",objective_percentile="{success_rate.value}",result="ok",service_name="autometrics"}} 0.0"""
        assert total_count_ok in data

        total_count_error = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_function",module="autometrics.test_decorator",objective_name="{objective_name}",objective_percentile="{success_rate.value}",result="error",service_name="autometrics"}} 0.0"""
        assert total_count_error in data

    @pytest.mark.asyncio
    async def test_initialize_counters_async(self):
        """This is a test to see if the function calls metric initializes at 0 after invoking the decorator for an async function"""

        autometrics(never_called_async_function)
        # NOTE - Do not call the function! We want to see if we get counter data for it even without ever calling it

        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count_ok = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_async_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="ok",service_name="autometrics"}} 0.0"""
        assert total_count_ok in data

        total_count_error = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_async_function",module="autometrics.test_decorator",objective_name="",objective_percentile="",result="error",service_name="autometrics"}} 0.0"""
        assert total_count_error in data

    @pytest.mark.asyncio
    async def test_initialize_counters_async_with_objective(self):
        """This is a test to see if the function calls metric initializes at 0 after invoking the decorator for an async function"""

        objective_name = "test_objective"
        success_rate = ObjectivePercentile.P90
        objective = Objective(name=objective_name, success_rate=success_rate)

        autometrics(objective=objective)(never_called_async_function)
        # NOTE - Do not call the function! We want to see if we get counter data for it even without ever calling it

        blob = generate_latest()
        assert blob is not None
        data = blob.decode("utf-8")

        total_count_ok = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_async_function",module="autometrics.test_decorator",objective_name="{objective_name}",objective_percentile="{success_rate.value}",result="ok",service_name="autometrics"}} 0.0"""
        assert total_count_ok in data

        total_count_error = f"""function_calls_total{{caller_function="",caller_module="",function="never_called_async_function",module="autometrics.test_decorator",objective_name="{objective_name}",objective_percentile="{success_rate.value}",result="error",service_name="autometrics"}} 0.0"""
        assert total_count_error in data
